
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2006 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <initng.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

#include <initng_global.h>
#include <initng_active_state.h>
#include <initng_active_db.h>
#include <initng_process_db.h>
#include <initng_service_cache.h>
#include <initng_handler.h>
#include <initng_active_db.h>
#include <initng_toolbox.h>
#include <initng_plugin_hook.h>
#include <initng_load_module.h>
#include <initng_plugin_callers.h>
#include <initng_error.h>
#include <initng_plugin.h>
#include <initng_static_states.h>
#include <initng_control_command.h>

#include <initng-paths.h>

#include "initng_nge.h"

/* globals */
struct stat sock_stat;
const char *socket_filename;

#define MAX_LISSENERS 20
int lisseners[MAX_LISSENERS + 1];

int is_active = FALSE;

void event_acceptor(f_module_h * from, e_fdw what);
static void close_all_lisseners(void);
static void close_initziator_socket(void);
static int open_initziator_socket(void);
static void check_socket(int signal);
void send_to_all(const void *buf, size_t len);


static int astatus_change(active_db_h * service);
static void system_state_change(e_is state);
static int system_pipe_watchers(active_db_h * service, process_h * process,
								char *output);
static int print_error(e_mt mt, const char *file, const char *func, int line,
					   const char *format, va_list arg);

/* todo, when last lissener closed, del hooks to save cpu cykles */

f_module_h fd_event_acceptor = { &event_acceptor, FDW_READ, -1 };

static void close_all_lisseners(void)
{
	int i;

	for (i = 0; i < MAX_LISSENERS; i++)
	{
		if (lisseners[i] > 0)
		{
			send(lisseners[i], "</disconnect>\n", sizeof(char) * 16, 0);

			close(lisseners[i]);
			lisseners[i] = -1;
		}
	}
}

static void close_initziator_socket(void)
{
	/* Check if we need to remove hooks */
	if (fd_event_acceptor.fds < 0)
		return;

	/* only remove if hook is added */
	if (is_active)
	{
		/*
		 * UnRegister that hooks, that we forwards events from.
		 */
		initng_plugin_hook_unregister(&g.ASTATUS_CHANGE, &astatus_change);
		initng_plugin_hook_unregister(&g.SWATCHERS, &system_state_change);
		initng_plugin_hook_unregister(&g.PIPEWATCHERS, &system_pipe_watchers);
		initng_plugin_hook_unregister(&g.ERR_MSG, &print_error);

		is_active = FALSE;
	}

	/* close socket and set to 0 */
	close(fd_event_acceptor.fds);
	fd_event_acceptor.fds = -1;

	/* remove tha hook too */
	initng_plugin_hook_unregister(&g.FDWATCHERS, &fd_event_acceptor);
}

/* send to all lisseners */
void send_to_all(const void *buf, size_t len)
{
	D_("send_to_all(%s)\n", (char *) buf);
	int i;

	/* walk all lissiners */
	for (i = 0; i < MAX_LISSENERS; i++)
	{
		/* if its not set */
		if (lisseners[i] < 1)
			continue;

		D_("Sending to lisseners[%i] fd %i : %s\n", i, lisseners[i],
		   (char *) buf);
		/* if not succed to send */
		if (send(lisseners[i], buf, len, 0) < (signed) len)
		{
			D_("Fd %i must have been closed.\n", lisseners[i]);
			/* close it */
			close(lisseners[i]);
			lisseners[i] = -1;
		}
	}

}

/* called by fd hook, when data is no socket */
void event_acceptor(f_module_h * from, e_fdw what)
{
	/* Temporary variables for sending data */
	char *string = NULL;
	int len;

	/* the fd the new lisseners will get */
	int lis = 0;

	/* make a dumb check */
	if (from != &fd_event_acceptor)
		return;

	/* skipp all set lisseners, so we dont owerwrite them */
	while (lisseners[lis] > 0 && lis != MAX_LISSENERS)
		lis++;
	if (lis == MAX_LISSENERS)
	{
		F_("Maximum no of lisseners reached.\n");
		return;
	}

	D_("Adding new lissener lissensers[%i]\n", lis);
	if (is_active == FALSE)
	{
		/*
		 * Register that hooks, that we forwards events from.
		 */
		initng_plugin_hook_register(&g.ASTATUS_CHANGE, 50, &astatus_change);
		initng_plugin_hook_register(&g.SWATCHERS, 50, &system_state_change);
		initng_plugin_hook_register(&g.PIPEWATCHERS, 50,
									&system_pipe_watchers);
		initng_plugin_hook_register(&g.ERR_MSG, 50, &print_error);
		is_active = TRUE;
	}
	/* create a new socket, for reading */
	if ((lisseners[lis] = accept(fd_event_acceptor.fds, NULL, NULL)) < 1)
	{
		F_("Failed to accept lissener!\n");
		return;
	}

	D_("opening lissener no #%i.\n", lis);

	/* send header */
	{
		send(lisseners[lis], "<? xml version=\"1.0\" ?/>\n",
			 sizeof(char) * 25, 0);
	}

	/* send protocol info */
	{
		string = i_calloc(70 + strlen(INITNG_VERSION), sizeof(char));

		/* Make a string ready for sending */
		len = sprintf(string,
					  "<connect protocol_version=\"%i\", initng_version=\"%s\"/>\n",
					  NGE_VERSION, INITNG_VERSION);

		/* send the init string to this socket */
		send(lisseners[lis], string, sizeof(char) * len, 0);

		/* free the initial string */
		free(string);
		string = NULL;
	}

	/* send system initiating state */
	{
		string = i_calloc(70, sizeof(char));
		len = sprintf(string,
					  "<event type=\"initial_system_state\" system_state=\"%i\" />\n",
					  g.sys_state);

		/* send the init string to this socket */
		send(lisseners[lis], string, sizeof(char) * len, 0);

		/* free the initial string */
		free(string);
		string = NULL;
	}

	/* send all current services states */
	{
		active_db_h *service = NULL;

		while_active_db(service)
		{
			string = i_realloc(string, (120 + strlen(service->name) +
										strlen(service->current_state->
											   state_name)) * sizeof(char));

			len = sprintf(string,
						  "<event type=\"initial_service_state\" service=\"%s\" is=\"%i\" state=\"%s\" />\n",
						  service->name, service->current_state->is,
						  service->current_state->state_name);

			/* send the init string to this socket */
			send(lisseners[lis], string, sizeof(char) * len, 0);
		}

		if (string)
		{
			/* free the initial string */
			free(string);
			string = NULL;
		}
	}

	/* tell client initialization is finished */
	{
		string = i_calloc(50, sizeof(char));
		len = sprintf(string, "<event type=\"initial_state_finished\" />\n");

		/* send the init string to this socket */
		send(lisseners[lis], string, sizeof(char) * len, 0);

		/* free the initial string */
		free(string);
		string = NULL;
	}

}


/* This will try to open a new socket, clients can iniziate to */
static int open_initziator_socket(void)
{
	/*    int flags; */
	struct sockaddr_un serv_sockname;

	/* Close the iniztiator */
	close_initziator_socket();

	/* Make NGE_PREFIX (/dev/initng) if it doesn't exist (try either way) */
	if (mkdir(NGE_PREFIX, S_IRUSR | S_IWUSR | S_IXUSR) == -1
		&& errno != EEXIST)
	{
		if (errno != EROFS)
			F_("Could not create " NGE_PREFIX
			   " : %s, may be / fs not mounted read-write yet?, will retry until I succeed.\n",
			   strerror(errno));
		return (FALSE);
	}

	/* chmod root path for root use only */
	if (chmod(NGE_PREFIX, S_IRUSR | S_IWUSR | S_IXUSR) == -1)
	{
		/* path doesn't exist, we don't have /dev yet. */
		if (errno == ENOENT || errno == EROFS)
			return (FALSE);

		F_("CRITICAL, failed to chmod %s, THIS IS A SECURITY PROBLEM.\n",
		   NGE_PREFIX);
	}

	/* Create the socket. */
	fd_event_acceptor.fds = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd_event_acceptor.fds < 1)
	{
		F_("Failed to init socket (%s)\n", strerror(errno));
		fd_event_acceptor.fds = -1;
		return (FALSE);
	}

	/* Bind a name to the socket. */
	serv_sockname.sun_family = AF_UNIX;

	/* get the socket_filename, set in module_init() */
	strcpy(serv_sockname.sun_path, socket_filename);

	/* Remove old socket file if any */
	unlink(serv_sockname.sun_path);

	/* Try to bind */
	if (bind
		(fd_event_acceptor.fds, (struct sockaddr *) &serv_sockname,
		 (strlen(serv_sockname.sun_path) +
		  sizeof(serv_sockname.sun_family))) < 0)
	{
		F_("Error binding to socket (errno: %d str: '%s')\n", errno,
		   strerror(errno));
		close_initziator_socket();
		unlink(serv_sockname.sun_path);
		return (FALSE);
	}

	/* chmod socket for root only use */
	if (chmod(serv_sockname.sun_path, S_IRUSR | S_IWUSR) == -1)
	{
		F_("CRITICAL, failed to chmod %s, THIS IS A SECURITY PROBLEM.\n",
		   serv_sockname.sun_path);
		close_initziator_socket();
		return (FALSE);
	}

	/* store sock_stat for checking if we need to recreate socket later */
	stat(serv_sockname.sun_path, &sock_stat);

	/* Listen to socket */
	if (listen(fd_event_acceptor.fds, 5))
	{
		F_("Error on listen (errno: %d str: '%s')\n", errno, strerror(errno));
		close_initziator_socket();
		unlink(serv_sockname.sun_path);
		return (FALSE);
	}

	/*
	 * Add an hook, so when fd_event_acceptor.fds have data,
	 * fd_event_acceptor.call (event_acceptor()) is called.
	 */
	initng_plugin_hook_register(&g.FDWATCHERS, 30, &fd_event_acceptor);

	/* return happily */
	return (TRUE);
}

/* this will check socket, and reopen on failure */
static void check_socket(int signal)
{
	struct stat st;

	if (signal != SIGHUP)
		return;

#define PING "<event type=\"ping\"/>\n"
	send_to_all(PING, sizeof(char) * strlen(PING));
	D_("Checking socket\n");

	/* Check if socket needs reopening */
	if (fd_event_acceptor.fds <= 0)
	{
		D_("fd_event_acceptor.fds not set, opening new socket.\n");
		open_initziator_socket();
		return;
	}

	/* stat the socket, reopen on failure */
	memset(&st, 0, sizeof(st));
	if (stat(socket_filename, &st) < 0)
	{
		W_("Stat failed! Opening new socket.\n");
		open_initziator_socket();
		return;
	}

	/* compare socket file, with the one that we know, reopen on failure */
	if (st.st_dev != sock_stat.st_dev || st.st_ino != sock_stat.st_ino
		|| st.st_mtime != sock_stat.st_mtime)
	{
		F_("Invalid socket found, reopening\n");
		open_initziator_socket();
		return;
	}

	D_("Socket ok.\n");
	return;
}

static int astatus_change(active_db_h * service)
{
	char *buffert = NULL;
	int len;

	buffert = calloc(120 + strlen(service->name) +
					 strlen(service->current_state->state_name),
					 sizeof(char));

	len = sprintf(buffert,
				  "<event type=\"service_state_change\" service=\"%s\" is=\"%i\" state=\"%s\" />\n",
				  service->name, service->current_state->is,
				  service->current_state->state_name);

	/*printf("astatus_change: %s %i %s\n",
	   service->name, service->current_state->is,
	   service->current_state->state_name); */

	if (len > 1)
		send_to_all(buffert, sizeof(char) * len);

	free(buffert);
	return (TRUE);
}

static void system_state_change(e_is state)
{
	char buffert[75];
	int len;


	len = sprintf(buffert,
				  "<event type=\"system_state_change\" system_state=\"%i\" />\n",
				  state);

	if (len > 1)
		send_to_all(buffert, sizeof(char) * len);
}

static int system_pipe_watchers(active_db_h * service, process_h * process,
								char *output)
{
	char *buffert = NULL;
	int len;
	buffert = calloc(100 + strlen(service->name) + strlen(process->pt->name) +
					 strlen(output), sizeof(char));

	len = sprintf(buffert,
				  "<event type=\"service_output\" service=\"%s\" process=\"%s\">%s</event>\n",
				  service->name, process->pt->name, output);

	if (len > 0)
		send_to_all(buffert, sizeof(char) * len);

	/* free buffert */
	free(buffert);

	/* return, output not handled */
	return (FALSE);
}

static int print_error(e_mt mt, const char *file, const char *func,
					   int line, const char *format, va_list arg)
{
	char *buffert = NULL;

	char *msg;
	int len, size;

	size = 256;
	msg = i_calloc(1, size);
	len = vsnprintf(msg, size, format, arg);
	while (len < 0 || len >= size)
	{
		/* Some glibc versions apparently return -1 if buffer too small.
		   Oh, and the argument counts the null char, but the return 
		   value doesn't, apparently. (See man page for gory details.) */
		size = (len < 0 ? size * 2 : len + 1);
		free(msg);
		msg = i_calloc(1, size);
		len = vsnprintf(msg, size, format, arg);
	}

	buffert = calloc(100 + len + strlen(file) + strlen(func), sizeof(char));

	len = sprintf(buffert,
				  "<event type=\"err_msg\" mt=\"%i\" file=\"%s\" func=\"%s\" line=\"%i\">%s</event>\n",
				  mt, file, func, line, msg);

	send_to_all(buffert, sizeof(char) * len);

	free(msg);
	free(buffert);

	return (FALSE);
}


int module_init(int api_version)
{
	int i;

	if (api_version != API_VERSION)
	{
		F_("This module is compiled for api_version %i version and initng is compiled on %i version, won't load this module!\n", API_VERSION, api_version);
		return (FALSE);
	}

	/* clear lisseners struct */
	for (i = 0; i < MAX_LISSENERS; i++)
		lisseners[i] = -1;

	/* zero globals */
	fd_event_acceptor.fds = -1;
	memset(&sock_stat, 0, sizeof(sock_stat));

	/*
	 * Decide witch path to socket to use.
	 * set socket_filename to right path
	 */

	/* decide which socket to use */
	if (g.i_am == I_AM_INIT)
		socket_filename = NGE_REAL;
	else if (g.i_am == I_AM_FAKE_INIT)
		socket_filename = NGE_TEST;
	else
		return (FALSE);
	D_("Socket is: %s\n", socket_filename);


	/*
	 * Giving initng a SIGHUP, will make initng check that all sockets are open,
	 * and reopen the sockets that have been deleted.
	 */
	initng_plugin_hook_register(&g.SIGNAL, 50, &check_socket);

	/* do the first socket directly */
	open_initziator_socket();


	/* return happily */
	return (TRUE);
}




void module_unload(void)
{

	/* close initziator socket */
	close_initziator_socket();

	/* dissconect all lisseners */
	close_all_lisseners();

	/* remove g.SIGNAL check hook */
	initng_plugin_hook_unregister(&g.SIGNAL, &check_socket);

}
